# Author: Cameron F. Abrams <cfa22@drexel.edu>

import pandas as pd
import numpy as np

import argparse as ap
from difflib import SequenceMatcher
from importlib.resources import files

from .compound import Compound

class PropertiesDatabase:
    resources_root = files('sandlerprops') / 'resources'
    """ Root directory for resource files. """
    data_dir = resources_root / 'data'
    """ Directory for data files. """
    datafile_path = data_dir / 'properties_database.csv'
    """ Path to the properties database CSV file. """
    def __init__(self):
        D = pd.read_csv(self.datafile_path, header=0, index_col=None)
        self.D = D.rename(columns={
            'Tfp (K)': 'Tfp',
            'Tb (K)': 'Tb',
            'Tc (K)': 'Tc',
            'Pc (bar)': 'Pc'})
        if self._swap_zeros_for_Os_in_Formulas() > 0:
            raise ValueError('Some formulas had 0 replaced with O; please verify correctness.')
        unitlist = [
            '', # no. (unique)
            '', # formula
            '', # name (unique)
            'g/mol', # molecular weight
            'K', # triple point temperature
            'K', # boiling point temperature
            'K', # critical temperature
            'bar', # critical pressure
            'm3/mol', # critical volume
            '', # critical compressibility
            '', # acentric factor
            '', # DIPM
            'J/mol-K', # ideal gas heat capacity coeff 1
            'J/mol-K2', # ideal gas heat capacity coeff 2
            'J/mol-K3', # ideal gas heat capacity coeff 3
            'J/mol-K4', # ideal gas heat capacity coeff 4
            'J/mol', # ideal gas enthalpy of formation at 298.15 K
            'J/mol', # ideal gas entropy of formation at 298.15 K   
            '', # vapor pressure equation type number
            '', # vapor pressure coeff 1
            '', # vapor pressure coeff 2
            '', # vapor pressure coeff 3
            '', # vapor pressure coeff 4
            'K', # vapor pressure temperature range min
            'K', # vapor pressure temperature range max
            '', # liquid density at Tden
            ''] # liquid density temperature for reference
        formatters = [
            '{:<10d}', # no. (unique)
            '{:<s}', # formula
            '{:<s}', # name (unique)
            '{:< 10.3f}', # molecular weight
            '{:< 10.1f}', # triple point temperature
            '{:< 10.1f}', # boiling point temperature
            '{:< 10.1f}', # critical temperature
            '{:< 10.2f}', # critical pressure
            '{:< 10.3f}', # critical volume
            '{:< 10.3f}', # critical compressibility
            '{:< 10.3f}', # acentric factor
            '{:<10g}', # DIPM
            '{:< 10.2f}', # ideal gas heat capacity coeff 1
            '{:< 10.4f}', # ideal gas heat capacity coeff 2
            '{:< 10.4e}', # ideal gas heat capacity coeff 3
            '{:< 10.4e}', # ideal gas heat capacity coeff 4
            '{:< 10.1f}', # ideal gas enthalpy of formation at 298.15 K
            '{:< 10.1f}', # ideal gas entropy of formation at 298.15 K   
            '{:<10d}', # vapor pressure equation type number
            '{:< 10.5f}', # vapor pressure coeff 1
            '{:< 10.5f}', # vapor pressure coeff 2
            '{:< 10.5f}', # vapor pressure coeff 3
            '{:< 10.5f}', # vapor pressure coeff 4
            '{:< 10.1f}', # vapor pressure temperature range min
            '{:< 10.1f}', # vapor pressure temperature range max
            '{:< 10.3f}', # liquid density at Tden
            '{:< 10.1f}'] # liquid density temperature for reference
        self.properties = list(self.D.columns)
        unitdict = {k: v for k,v in zip(self.properties, unitlist)}
        formatterdict = {k: v for k,v in zip(self.properties, formatters)}
        self.U = ap.Namespace(**unitdict)
        self.F = ap.Namespace(**formatterdict)

    def _swap_zeros_for_Os_in_Formulas(self):
        """
        Internal method to replace zeros with capital 'O's in the Formula column of the DataFrame.
        However, we can only do this if either of these conditions are met:
        1. The zero is the first character in the Formula
        2. The zero is preceded by a letter (not another digit)
        """
        num_corrections = 0
        formulas = self.D['Formula'].to_list()
        for i, formula in enumerate(formulas):
            trigger = False
            new_formula_chars = []
            for j, char in enumerate(formula):
                if char == '0':
                    if j == 0 or (j > 0 and not formula[j-1].isdigit()):
                        trigger = True
                        new_formula_chars.append('O')
                    else:
                        new_formula_chars.append('0')
                else:
                    new_formula_chars.append(char)
            new_formula = ''.join(new_formula_chars)
            if trigger:
                num_corrections += 1
                print(f'Corrected Formula from {formula} to {new_formula}')
            self.D.at[i, 'Formula'] = new_formula
        return num_corrections
    
    def show_properties(self, args: ap.Namespace = None):
        """ 
        Subcommand handler that displays the list of available properties with their units.
        
        Parameters
        ----------
        args : argparse.Namespace, optional
            Not used; present for compatibility since this is a subcommand handler.
        """
        for p in self.properties:
            unit = self.U.__dict__[p]
            if unit:
                fmted = self.F.__dict__[p].format(p)
                print(f'{fmted} ({unit})')
            else:
                print(f'{fmted}')

    def find_compound(self, args: ap.Namespace):
        """
        Subcommand handler that looks for a compound by name and displays if found.
        
        Parameters
        ----------
        args : argparse.Namespace
            Must contain attribute 'compound_name' with the name of the compound to find.
        """
        compound_name = args.compound_name
        record = self.get_compound(compound_name)
        if record is not None:
            print(f'Found exact match: {record.Name} (index {record.No})')

    def show_compound_properties(self, args: ap.Namespace):
        """
        Subcommand handler that displays all properties of a specified compound.
        
        Parameters
        ----------
        args : argparse.Namespace
            Must contain attribute 'compound_name' with the name of the compound to display.    
        """
        compound_name = args.compound_name
        record = self.get_compound(compound_name)
        if record is not None:
            print(f'Properties of {record.Name} (index {record.No}):')
            for p in self.properties:
                value = record.__dict__[p]
                unit = self.U.__dict__[p]
                formatted_value = self.F.__dict__[p].format(value)
                if unit:
                    print(f'  {p:<10s}: {formatted_value} {unit}')
                else:
                    print(f'  {p:<10s}: {formatted_value}')

    def get_compound(self, name: str, near_matches: int = 10):
        """ 
        Retrieves a **Compound** by name. If not found, suggests similar names and returns
        an unpopulated **Compound** object with name and Formula set to input name.
        
        Parameters
        ----------
        name : str
            Name of the compound to retrieve
        near_matches : int
            Number of similar names to suggest if exact match not found
        
        Returns
        -------
        **Compound**
            The **Compound** object if found, else an empty **Compound** with name and Formula set.
        """
        row = self.D[self.D['Name'] == name]
        if not row.empty:
            d = row.to_dict('records')[0]
            return Compound(**d)
        else:
            print(f'{name} not found; empty Compound object returned.  Here are similars:')
            scores = []
            for n in self.D['Name']:
                scores.append(SequenceMatcher(None, name, n).ratio())
            scores = np.array(scores)
            si = np.argsort(scores)
            sorted_names = np.array(self.D['Name'])[si]
            top_sorted_names = sorted_names[-near_matches:][::-1]
            for n in top_sorted_names:
                print(n)
        return Compound(Name=name, Formula=name)


_instance = None

def get_database():
    """Get or create the singleton database instance."""
    global _instance
    if _instance is None:
        _instance = PropertiesDatabase()
    return _instance


# Author: Cameron F. Abrams, <cfa22@drexel.edu>
import argparse as ap

from .properties import PropertiesDatabase

banner = """
                                  █████ ████                    
                                 ░░███ ░░███                    
  █████   ██████   ████████    ███████  ░███   ██████  ████████ 
 ███░░   ░░░░░███ ░░███░░███  ███░░███  ░███  ███░░███░░███░░███
░░█████   ███████  ░███ ░███ ░███ ░███  ░███ ░███████  ░███ ░░░ 
 ░░░░███ ███░░███  ░███ ░███ ░███ ░███  ░███ ░███░░░   ░███     
 ██████ ░░████████ ████ █████░░████████ █████░░██████  █████    
░░░░░░   ░░░░░░░░ ░░░░ ░░░░░  ░░░░░░░░ ░░░░░  ░░░░░░  ░░░░░     
                                                                                 
       ████████  ████████   ██████  ████████   █████            
      ░░███░░███░░███░░███ ███░░███░░███░░███ ███░░             
       ░███ ░███ ░███ ░░░ ░███ ░███ ░███ ░███░░█████            
       ░███ ░███ ░███     ░███ ░███ ░███ ░███ ░░░░███           
       ░███████  █████    ░░██████  ░███████  ██████            
       ░███░░░  ░░░░░      ░░░░░░   ░███░░░  ░░░░░░             
       ░███                         ░███                        
       █████                        █████                       
      ░░░░░                        ░░░░░              
        (c) 2025, Cameron F. Abrams <cfa22@drexel.edu> 
"""

def cli():
    P = PropertiesDatabase()
    subcommands = {
        'showprops': dict(
            func = P.show_properties,
            help = 'show available properties',
            ),
        'find' : dict(
            func = P.find_compound,
            help = 'find compound by name',
            ),
        'show': dict(
            func = P.show_compound_properties,
            help = 'show properties for a compound',
        )
    }
    parser = ap.ArgumentParser(
        prog='sandlerprops',
        description='sandlerprops -- Sandler Properties Database CLI',
        epilog='(c) 2025 Cameron F. Abrams <cfa22@drexel.edu>'
    )
    parser.add_argument(
        '-b',
        '--banner',
        default=False,
        action=ap.BooleanOptionalAction,
        help='toggle banner message'
    )
    subparsers = parser.add_subparsers(
        title="subcommands",
        dest="command",
        metavar="<command>",
        required=True,
    )
    command_parsers={}
    for k, specs in subcommands.items():
        command_parsers[k] = subparsers.add_parser(
            k,
            help=specs['help'],
            formatter_class=ap.RawDescriptionHelpFormatter
        )
        command_parsers[k].set_defaults(func=specs['func'])

    command_parsers['find'].add_argument(
        'compound_name',
        type=str,
        help='name of compound to find'
    )
    command_parsers['show'].add_argument(
        'compound_name',
        type=str,
        help='name of compound whose properties to show'
    )

    args = parser.parse_args()
    if args.banner:
        print(banner)
    if hasattr(args, 'func'):
        args.func(args)
    else:
        my_list = ', '.join(list(subcommands.keys()))
        print(f'No subcommand found. Expected one of {my_list}')
    if args.banner:
        print('Thanks for using sandlerprops!')


import numpy as np
from dataclasses import dataclass, field

@dataclass
class Compound:
    No: int = 0
    """ Unique compound number """
    Formula: str = ''
    """ Empirical formula """
    Name: str = ''
    """ Unique compound name """
    Molwt: float = 0.0
    """ Molecular weight in g/mol """
    Tfp: float = 0.0
    """ Triple point temperature in K """
    Tb: float = 0.0
    """ Boiling point temperature in K """
    Tc: float = 0.0
    """ Critical temperature in K """
    Pc: float = 0.0
    """ Critical pressure in **bar** """
    Vc: float = 0.0
    """ Critical volume in m3/mol """
    Zc: float = 0.0
    """ Critical compressibility """
    Omega: float = 0.0
    """ Acentric factor """
    Dipm: float = 0.0
    """ DIPM """
    CpA: float = 0.0
    """ Ideal gas heat capacity coeff 1 """
    CpB: float = 0.0
    """ Ideal gas heat capacity coeff 2 """
    CpC: float = 0.0
    """ Ideal gas heat capacity coeff 3 """
    CpD: float = 0.0
    """ Ideal gas heat capacity coeff 4 """
    dHf: float = 0.0
    """ Ideal gas enthalpy of formation at 298.15 K """
    dGf: float = 0.0
    """ Ideal gas entropy of formation at 298.15 K   """
    Eq: int = 0
    """ Vapor pressure equation type number """
    VpA: float = 0.0
    """ Vapor pressure coeff 1 """
    VpB: float = 0.0
    """ Vapor pressure coeff 2 """
    VpC: float = 0.0
    """ Vapor pressure coeff 3 """
    VpD: float = 0.0
    """ Vapor pressure coeff 4 """
    Tmin: float = 0.0
    """ Vapor pressure temperature range min """
    Tmax: float = 0.0
    """ Vapor pressure temperature range max """
    Lden: float = 0.0
    """ Liquid density at Tden """
    Tden: float = 0.0
    """ Temperature at which liquid density is measured """
    charge: int = 0
    """ Net charge of the compound (not in input properties set unless encoded in Formula) """
    atomset: set = field(default_factory=set)
    """ Set of unique atom names in empirical formula """
    atomdict: dict = field(default_factory=dict)
    """ Dictionary of atomname:count items representing empirical formula """

    def __post_init__(self):
        """ dictionary of atomname:count items representing empirical formula """
        efc = self.Formula.split('^')
        ef_neutral = efc[0]
        self.charge = 0
        if len(efc) > 1:
            expo = efc[1]
            if expo[0] == '{' and expo[-1] == '}':
                self.charge = int(expo[1:-1])
            else:
                raise ValueError(f'Error: malformed charge {efc[1]}')

        self.atomdict = parse_empirical_formula(ef_neutral)
        self._reorder_elements()
        self.atomset = set(self.atomdict.keys())

    @property
    def Cp(self):
        """ Returns ideal gas heat capacity coefficients as a numpy array """
        return np.array([self.CpA, self.CpB, self.CpC, self.CpD])

    def _reorder_elements(self):
        leave_these_alone = ['NH3', 'CH3OH', 
                             'Fe','He','Li','Be','Ne','Na','Mg','Al','Si','Cl','Ar','Ca','Sc',
                             'Ti','Cr','Mn','Co','Ni','Cu','Zn','Ga','Ge','As','Se','Br','Kr',
                             'Rb','Sr','Y','Zr','Nb','Mo','Tc','Ru','Rh','Pd','Ag','Cd','In',
                             'Sn','Sb','Te','I','Xe','Cs','Ba','La','Ce','Pr','Nd','Pm','Sm',
                             'Eu','Gd','Tb','Dy','Ho','Er','Tm','Yb','Lu','Hf','Ta','W','Re',
                             'Os','Ir','Pt','Au','Hg','Tl','Pb','Bi','Po','At','Rn','Fr','Ra',
                             'Ac','Th','Pa','U']
        if self.Formula in leave_these_alone:
            return
        my_order_preference = ['C', 'H', 'O', 'N', 'Na', 'K', 'Ca', 'F', 'Cl', 'Br', 'I']
        A = self.atomdict.copy()
        ef = ''
        for a in my_order_preference:
            if a in A:
                c = '' if A[a] == 1 else str(A[a])
                ef += f'{a}{c}'
                del A[a]
        for e, c in A.items():
            c = '' if c == 1 else str(c)
            ef += f'{e}{c}'
        self.Formula = ef

''' a bunch of functions that permit conversion of an empirical formula into
    an element:count dictionary '''
# per https://stackoverflow.com/users/5079316/olivier-melan%c3%a7on

def _push(obj,l,depth):
    while depth:
        l = l[-1]
        depth -= 1
    l.append(obj)

def _parse_parentheses(s):
    ''' byte-wise de-nestify a string with parenthesis '''
    groups = []
    depth = 0
    try:
        i=0
        while i<len(s):
            char=s[i]
            if char == '(':
                _push([], groups, depth)
                depth += 1
            elif char == ')':
                depth -= 1
            else:
                _push(char, groups, depth)
            i+=1
    except IndexError:
        raise ValueError('Parentheses mismatch')
    if depth != 0:
        raise ValueError('Parentheses mismatch')
    else:
        return groups

def bankblock(B,b):
    if len(b[0])>0: # bank this block
        if not any(isinstance(i, list) for i in b[0]):
            b[0]=''.join(b[0])
        nstr=''.join(b[1])
        b[1]=1 if len(nstr)==0 else int(nstr)
        B.append(b)

def blockify(bl):
    ''' parse the byte_levels returned from the byte-wise de-nester into blocks, where
        a block is a two-element list, where first element is a block and second is 
        an integer subscript >= 1.  A "primitive" block is one in which the first
        element is not a list, but instead a string that indentifies a chemical element. '''
    blocks=[]
    curr_block=[[],[]]
    for b in bl:
        if len(b)==1:
            if b.isalpha():
                if b.isupper(): # new block
                    bankblock(blocks,curr_block)
                    curr_block=[[b],[]]
                else: # still building this block's elem name
                    curr_block[0].append(b)
            elif b.isdigit():
                curr_block[1].append(b)
        else:
            bankblock(blocks,curr_block)
            curr_block=[blockify(b),[]]
    bankblock(blocks,curr_block)
    return blocks

def flattify(B):
    ''' distribute the block counts inward '''
    for b in B:
        if isinstance(b[0],str) or b[1]==1: # already flat
            pass
        else:
            m=b[1]
            b[1]=1
            for bb in b[0]:
                bb[1]*=m
                flattify(b[0])

def my_flatten(L,size=(2)):
    ''' flatten '''
    flatlist=[]
    for i in L:
        if not isinstance(i[0],list):
            flatlist.append(i)
        else:
            newlist=my_flatten(i[0])
            flatlist.extend(newlist)
    return flatlist

def reduce(L):
    ''' produce a dictionary of element:number '''
    result_dict={}
    for i in L:
        if i[0] in result_dict:
            result_dict[i[0]]+=i[1]
        else:
            result_dict[i[0]]=i[1]
    return result_dict

def parse_empirical_formula(ef):
    block_levels=blockify(_parse_parentheses(ef))
    flattify(block_levels)
    return reduce(my_flatten(block_levels))
